import collections
import os
from shutil import copyfile

import requests

import castero
from castero.net import Net


class DataFile:
    """The DataFile class.

    Used when handling files with data that can reasonably be stored in a
    dictionary. Particularly used in the Config class and the Feeds class.

    Extended by classes which are based on a data file.
    """
    PACKAGE = os.path.dirname(__file__)
    HOME = os.path.expanduser('~')
    XDG_CONFIG_HOME = os.getenv('XDG_CONFIG_HOME',
                                os.path.join(HOME, '.config'))
    XDG_DATA_HOME = os.getenv('XDG_DATA_HOME',
                              os.path.join(HOME, '.local', 'share'))
    CONFIG_DIR = os.path.join(XDG_CONFIG_HOME, castero.__title__)
    DATA_DIR = os.path.join(XDG_DATA_HOME, castero.__title__)
    DEFAULT_DOWNLOADED_DIR = os.path.join(DATA_DIR, "downloaded")

    def __init__(self, path, default_path) -> None:
        """Initializes the object.

        Args:
            path: the path to the data file
            default_path: the path to the default data file
        """
        assert os.path.exists(default_path)

        self.data = collections.OrderedDict()
        self._path = path
        self._default_path = default_path

        # if path doesn't exit, create it based on default_path
        if not os.path.exists(self._path):
            DataFile.ensure_path(self._path)
            copyfile(self._default_path, self._path)

    def __iter__(self) -> iter:
        """Iterator for the keys of self.data

        In order to iterate over data values, you should use something like:
        for key in file_instance:
            value = file_instance[key]
        """
        return self.data.__iter__()

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        self.data.pop(key, None)

    @staticmethod
    def ensure_path(filename):
        """Ensure that the path to the filename exists, creating it if needed.
        """
        path = os.path.dirname(filename)
        if not os.path.exists(path):
            os.makedirs(path)

    @staticmethod
    def download_to_file(url, file, name, download_queue, display=None):
        """Downloads a URL to a local file.

        Args:
            url: the source url
            file: the destination path
            name: the user-friendly name of the content
            download_queue: the download_queue overseeing this download
            display: (optional) the display to write status updates to
        """
        chunk_size = 1024
        chuck_size_label = "KB"

        response = Net.Get(url, stream=True)
        handle = open(file, "wb")
        downloaded = 0
        for chunk in response.iter_content(chunk_size=chunk_size):
            if display is not None:
                status_str = "Downloading \"%s\": %d%s" % (
                    name, downloaded / chunk_size, chuck_size_label
                )
                if download_queue.length > 1:
                    status_str += " (+%d downloads in queue)" % \
                                  (download_queue.length - 1)

                display.change_status(status_str)
            if chunk:
                handle.write(chunk)
            downloaded += len(chunk)

        if display is not None:
            display.change_status("Episode successfully downloaded.")
        download_queue.next()

    def load(self) -> None:
        """Loads the data file.

        Should be implemented by classes which extend this class.
        """
        pass

    def write(self) -> None:
        """Writes to the data file.

        Should be implemented by classes which extend this class.
        """
        pass



from bs4 import BeautifulSoup
import re


def third(n) -> int:
    """Calculates one-third of a given value.

    Args:
        n: the integer to calculate one-third of

    Returns:
        int: one-third of n, rounded down
    """
    return int(n / 3)


def median(arr):
    """Determines the median of a list of numbers.

    Args:
        arr: a list of ints and/or floats of which to determine the median

    Returns:
        int or float: the median value in arr
    """
    if len(arr) == 0:
        return None

    arr_sorted = sorted(arr)
    midpoint = int(len(arr) / 2)

    if len(arr) % 2 == 0:
        # even number of elements; get the average of the middle two
        result = (arr_sorted[midpoint - 1] + arr_sorted[midpoint]) / 2
    else:
        result = arr_sorted[midpoint]
    return result


def sanitize_path(path) -> str:
    """Replaces any characters in path that the file system may not support.

    This method replaces any non-alphanumeric characters with an underscore,
    with the exception of hyphens.

    Args:
        path: the original path

    Returns:
        str: the given path with potentially unsafe characters replaced
    """
    # adapted from https://stackoverflow.com/a/13593932
    path = re.sub('[^\w\-]', '_', path)
    return path


def is_true(string) -> bool:
    """Determines whether a string represents True.

    As the name suggests, any input which is not explicitly evaluated to True
    will cause this method to return False.

    Args:
        string: the string to evaluate
    """
    assert isinstance(string, str)

    return string in ['True', 'true', '1']


def html_to_plain(html) -> str:
    """Converts a potentially HTML-formatted string to user-friendly plaintext.
    
    Args:
        html: the text to convert with potential html tags

    Returns:
        str: the given text with html tags removed
    """
    soup = BeautifulSoup(html, 'html.parser')
    return soup.get_text()


import configparser
import os
import sys

from castero.datafile import DataFile


class ConfigError(Exception):
    """An ambiguous error while handling the configuration.
    """


class ConfigParseError(ConfigError):
    """An error occurred while parsing the config file.
    """


class ConfigDuplicateError(ConfigError):
    """The config file contained a duplicate variable.
    """


class _Config(DataFile):
    """The Config class.

    Reads the configuration file. Instances of this class can generally be
    treated like dictionaries, accessing a variable with config_instance[key].

    Modifying config variables inside the application is not supported; config
    changes must be made to the config file itself.
    """
    PATH = os.path.join(DataFile.CONFIG_DIR, 'castero.conf')
    DEFAULT_PATH = os.path.join(DataFile.PACKAGE, 'templates/castero.conf')

    def __init__(self) -> None:
        """Initializes the object.
        """
        super().__init__(self.PATH, self.DEFAULT_PATH)

        # strictly use default path when testing
        if "pytest" in sys.modules:
            self._path = self._default_path

        self.load()

    def __setitem__(self, key, value):
        pass

    def load(self) -> None:
        """Loads the config file.

        Raises:
            ConfigParseError: an error occurred while parsing the config file
        """
        assert os.path.exists(self._path)
        assert os.path.exists(self._default_path)

        conf = configparser.ConfigParser()
        try:
            conf.read(self._path)
        except configparser.ParsingError:
            raise ConfigParseError(
                "An error occurred while parsing the config file"
            )

        # we also read from the the default_conf to make sure conf contains
        # all of the necessary parameters
        default_conf = configparser.ConfigParser()
        try:
            default_conf.read(self._default_path)
        except configparser.ParsingError:
            raise ConfigParseError(
                "An error occurred while parsing the default config file"
                " (don't modify that one!)"
            )

        # ensure that all variables in the default_conf are present
        for section in default_conf:
            if section not in conf:
                self.migrate(conf, default_conf)
                break  # we will only ever need to migrate once
            for key in default_conf[section]:
                if key not in conf[section]:
                    self.migrate(conf, default_conf)
                    break  # we will only ever need to migrate once

        for section in conf:
            for key in conf[section]:
                if section in default_conf:
                    if key in default_conf[section]:
                        # sections in the config file are purely for aesthetic
                        # purposes - this config object only stores variables
                        # at a single depth
                        if key in self.data:
                            raise ConfigDuplicateError(
                                "Variable defined multiple times, key: " + key
                            )

                        self.data[key] = conf[section][key]
                    else:
                        # disallow keys which are not in the default config
                        self.migrate(conf, default_conf)
                else:
                    self.migrate(conf, default_conf)

    def migrate(self, conf, default_conf) -> None:
        """Migrates the user's config to have the same keys as the default.

        Generally this method will be used when the client is updated to change
        some parts of the default config.

        The migration process is rather brash -- it simply retains the user's
        current config in memory, overwrites their file with the default file,
        and then using basic parsing replaces any variables already defined
        by the user with their definition.

        Args:
            conf: ConfigParser of the user's config file
            default_conf: ConfigParser of the default config file
        """
        # convert conf and default_conf to 1-dim dictionaries since they may
        # not have the same sections
        conf_dict = {}
        for section in conf:
            for key in conf[section]:
                conf_dict[key] = conf[section][key]
        default_conf_dict = {}
        for section in default_conf:
            for key in default_conf[section]:
                default_conf_dict[key] = default_conf[section][key]

        with open(self._default_path, "r") as default_conf_file:
            lines = default_conf_file.readlines()
            for line in lines:
                for key in default_conf_dict:
                    if line.startswith(key + " "):
                        if key in conf_dict:
                            lines[lines.index(line)] = "%s = %s\n" % (
                                line.split(" =")[0],
                                conf_dict[key]
                            )

        with open(self._path, "w") as conf_file:
            for line in lines:
                conf_file.write(line)
        conf.read(self._path)


Config = _Config()



import curses
import sys
import threading
import re

import castero
from castero import helpers
from castero.config import Config
from castero.database import Database
from castero.display import Display
from castero.player import Player


def main():
    database = Database()

    # update fields in help menu text
    for field in Config:
        if "{%s}" % field in castero.__help__:
            castero.__help__ = \
                castero.__help__.replace(
                    "{%s}" % field,
                    Config[field].ljust(9)
                )
        elif "{%s|" % field in castero.__help__:
            field2 = castero.__help__.split("{%s|" % field)[1].split("}")[0]
            castero.__help__ = \
                castero.__help__.replace(
                    "{%s|%s}" % (field, field2),
                    ("%s or %s" % (Config[field], Config[field2])).ljust(9)
                )
    remaining_brace_fields = re.compile('\{.*?\}').findall(castero.__help__)
    for field in remaining_brace_fields:
        adjusted = field.replace("{", "").replace("}", "").ljust(9)
        castero.__help__ = \
            castero.__help__.replace(field, adjusted)

    # check if user is running the client with an info flag
    info_flags = {
        'help': ['-h', '--help'],
        'version': ['-v', '--version']
    }
    if sys.argv[len(sys.argv) - 1] in info_flags['help']:
        print(castero.__help__)
        sys.exit(0)
    elif sys.argv[len(sys.argv) - 1] in info_flags['version']:
        print(castero.__version__)
        sys.exit(0)

    # check whether dependencies are met
    Player.check_dependencies()

    # instantiate the display object
    stdscr = curses.initscr()
    display = Display(stdscr, database)
    display.clear()
    display.update_parent_dimensions()

    # check if we need to start reloading
    if helpers.is_true(Config['reload_on_start']):
        reload_thread = threading.Thread(target=feeds.reload, args=[display])
        reload_thread.start()

    # run initial display operations
    display.display()
    display.update()
    display.refresh()

    # core loop for the client
    running = True
    while running:
        display.display()
        display.update()
        display.refresh()
        char = display.getch()
        if char != -1:
            running = display.handle_input(char)

    sys.exit(0)


main()



import json
import os
import sqlite3
from typing import List, Tuple

from castero.datafile import DataFile
from castero.episode import Episode
from castero.feed import Feed


class Database():
    PATH = os.path.join(DataFile.DATA_DIR, 'castero.db')
    OLD_PATH = os.path.join(DataFile.DATA_DIR, 'feeds')
    MIGRATIONS_DIR = os.path.join(DataFile.PACKAGE, 'templates/migrations')

    def __init__(self):
        existed = os.path.exists(self.PATH)
        self._conn = sqlite3.connect(self.PATH, check_same_thread=False)
        self._conn.execute("PRAGMA foreign_keys = ON")

        if not existed and os.path.exists(self.OLD_PATH):
            self._create_from_old_feeds()

        self.migrate()

    def migrate(self):
        cursor = self._conn.cursor()
        cur_version = cursor.execute('pragma user_version').fetchone()[0]

        migration_files = list(os.listdir(self.MIGRATIONS_DIR))
        for migration in sorted(migration_files):
            version = int(migration.split("-")[0])
            if version > cur_version:
                path = os.path.join(self.MIGRATIONS_DIR, migration)
                with open(path, 'rt') as f:
                    cursor.executescript(f.read())

    def _create_from_old_feeds(self):
        self.migrate()
        cursor = self._conn.cursor()

        with open(self.OLD_PATH, 'r') as f:
            content = json.loads(f.read())

        for key in content:
            feed_dict = content[key]

            sql = "insert into feed (key, title, description, link, last_build_date, copyright)\n" \
                  "values (?,?,?,?,?,?)"
            cursor.execute(sql, (
                key,
                feed_dict["title"],
                feed_dict["description"],
                feed_dict["link"],
                feed_dict["last_build_date"],
                feed_dict["copyright"]
            ))

            for episode_dict in feed_dict["episodes"]:
                sql = "insert into episode (feed_key, title, description, link, pubdate, copyright, enclosure)\n" \
                      "values (?,?,?,?,?,?,?)"
                cursor.execute(sql, (
                    key,
                    episode_dict["title"],
                    episode_dict["description"],
                    episode_dict["link"],
                    episode_dict["pubdate"],
                    episode_dict["copyright"],
                    episode_dict["enclosure"]
                ))

        self._conn.commit()

    def delete_feed(self, feed: Feed) -> None:
        sql = "delete from feed where key=?"
        cursor = self._conn.cursor()
        cursor.execute(sql, (feed.key,))
        self._conn.commit()

    def replace_feed(self, feed: Feed) -> None:
        sql = "replace into feed (key, title, description, link, last_build_date, copyright)\n" \
              "values (?,?,?,?,?,?)"
        cursor = self._conn.cursor()
        cursor.execute(sql, (
            feed.key,
            feed.title,
            feed.description,
            feed.link,
            feed.last_build_date,
            feed.copyright
        ))
        self._conn.commit()

    def replace_episode(self, feed: Feed, episode: Episode) -> None:
        cursor = self._conn.cursor()
        if episode.ep_id is None:
            sql = "replace into episode (title, feed_key, description, link, pubdate, copyright, enclosure, played)\n" \
                "values (?,?,?,?,?,?,?,?)"
            cursor.execute(sql, (
                episode.title,
                feed.key,
                episode.description,
                episode.link,
                episode.pubdate,
                episode.copyright,
                episode.enclosure,
                episode.played
            ))
            episode.ep_id = cursor.lastrowid
        else:
            sql = "replace into episode (id, title, feed_key, description, link, pubdate, copyright, enclosure, played)\n" \
                "values (?,?,?,?,?,?,?,?,?)"
            cursor.execute(sql, (
                episode.ep_id,
                episode.title,
                feed.key,
                episode.description,
                episode.link,
                episode.pubdate,
                episode.copyright,
                episode.enclosure,
                episode.played
            ))
        self._conn.commit()

    def replace_episodes(self, feed: Feed, episodes: List[Episode]) -> None:
        for episode in episodes:
            self.replace_episode(feed, episode)
        self._conn.commit()

    def feeds(self) -> List[Feed]:
        sql = "select key, title, description, link, last_build_date, copyright from feed"
        cursor = self._conn.cursor()
        cursor.execute(sql)

        feeds = []
        for row in cursor.fetchall():
            feeds.append(Feed(
                url=row[0] if row[0].startswith('http') else None,
                file=row[0] if not row[0].startswith('http') else None,
                title=row[1],
                description=row[2],
                link=row[3],
                last_build_date=row[4],
                copyright=row[5]
            ))
        return feeds

    def episodes(self, feed: Feed) -> List[Episode]:
        sql = "select id, title, description, link, pubdate, copyright, enclosure, played from episode where feed_key=?"
        cursor = self._conn.cursor()
        cursor.execute(sql, (feed.key,))

        episodes = []
        for row in cursor.fetchall():
            episodes.append(Episode(
                feed,
                ep_id=row[0],
                title=row[1],
                description=row[2],
                link=row[3],
                pubdate=row[4],
                copyright=row[5],
                enclosure=row[6],
                played=row[7]
            ))
        return episodes

    def feed(self, key) -> Feed:
        sql = "select key, title, description, link, last_build_date, copyright from feed where key=?"
        cursor = self._conn.cursor()
        cursor.execute(sql, (key,))

        result = cursor.fetchone()
        if result is None:
            return None
        else:
            return Feed(
                url=result[0] if result[0].startswith('http') else None,
                file=result[0] if not result[0].startswith('http') else None,
                title=result[1],
                description=result[2],
                link=result[3],
                last_build_date=result[4],
                copyright=result[5],
                episodes=[],
            )

    def episode(self, ep_id: int) -> Episode:
        sql = "select feed_key, id, title, description, link, pubdate, copyright, enclosure, played from episode where id=?"
        cursor = self._conn.cursor()
        cursor.execute(sql, (ep_id,))

        result = cursor.fetchone()
        if result is None:
            return None
        else:
            feed = self.feed(result[0])
            return Episode(
                feed,
                ep_id=result[1],
                title=result[2],
                description=result[3],
                link=result[4],
                pubdate=result[5],
                copyright=result[6],
                enclosure=result[7],
                played=result[8]
            )

    def reload(self, display=None) -> None:
        feeds = self.feeds()
        total_feeds = len(feeds)
        current_feed = 1

        for feed in self.feeds():
            if display is not None:
                display.change_status(
                    "Reloading feeds (%d/%d)" % (current_feed, total_feeds)
                )

            # assume urls have http in them
            if "http" in feed.key:
                new_feed = Feed(url=feed.key)
            else:
                new_feed = Feed(file=feed.key)

            self.replace_feed(new_feed)
            self.replace_episodes(new_feed, new_feed.parse_episodes())

        if display is not None:
            display.change_status("Feeds successfully reloaded")
            display.menus_valid = False


import curses

from castero import helpers
from castero.config import Config
from castero.episode import Episode
from castero.feed import Feed
from castero.menu import Menu


class EpisodeMenu(Menu):
    def __init__(self, window, source, child=None, active=False) -> None:
        super().__init__(window, source, child=child, active=active)

        self._feed = None
        self._episodes = []

    def __len__(self) -> int:
        return len(self._episodes)

    def _items(self):
        return [
            {
                'attr': curses.color_pair(5) if episode.played else curses.A_NORMAL,
                'text': str(episode)
            }
            for episode in self._episodes
        ]

    def item(self) -> Episode:
        if len(self._episodes) == 0:
            return None

        return self._episodes[self._selected]

    def metadata(self):
        if len(self._episodes) == 0:
            return ""

        return self._episodes[self._selected].metadata

    def update_items(self, feed: Feed):
        assert isinstance(feed, Feed) or feed is None

        super().update_items(feed)

        self._feed = feed

        if feed is None:
            self._episodes = []
        else:
            self._episodes = \
                [episode for episode in self._source.episodes(feed)]

            if self._inverted:
                self._episodes.reverse()

        self._sanitize()

    def update_child(self):
        pass

    def invert(self):
        super().invert()

        self.update_items(self._feed)



import castero
import requests


class Net:
    """The Net class.

    This class provides helper methods for network requests. Generally just a
    wrapper around the requests library.
    """
    USER_AGENT = "%s %s <%s>" % (
        castero.__title__, castero.__version__, castero.__url__)
    HEADERS = {
        'User-Agent': USER_AGENT
    }

    @staticmethod
    def Get(url, **kwargs) -> requests.models.Response:
        """Send a GET request.

        Args:
            url: URL to retrieve from
            **kwargs: optional arguments for requests.get()

        Returns:
            requests.models.Response: response
        """
        return requests.get(url, headers=Net.HEADERS, **kwargs)



import time

from castero.player import Player, PlayerDependencyError


class VLCPlayer(Player):
    """The VLCPlayer class.
    """
    NAME = "vlc"

    def __init__(self, title, path, episode) -> None:
        """Initializes the object.

        Overrides method from Player; see documentation in that class.
        """
        super().__init__(title, path, episode)

        import vlc
        self.vlc = vlc

    @staticmethod
    def check_dependencies():
        """Checks whether dependencies are met for playing a player.

        Overrides method from Player; see documentation in that class.
        """
        try:
            import vlc
            vlc.Instance()
        except (ImportError, NameError, OSError, AttributeError,
                ModuleNotFoundError):
            raise PlayerDependencyError(
                "Dependency VLC not found, which is required for playing"
                " media files"
            )

    def _create_player(self) -> None:
        """Creates the player object while making sure it is a valid file.

        Overrides method from Player; see documentation in that class.
        """
        vlc_instance = self.vlc.Instance("--no-video --quiet")

        self._player = vlc_instance.media_player_new()
        self._media = vlc_instance.media_new(self._path)
        self._media.parse()  # may output some junk into the console
        self._player.set_media(self._media)

        self._duration = self._media.get_duration()

    def play(self) -> None:
        """Plays the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is None:
            self._create_player()

        self._player.play()
        self._state = 1

    def stop(self) -> None:
        """Stops the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is not None:
            if self._player.get_state() == self.vlc.State.Opening:
                self._player.release()
            else:
                self._player.stop()
                self._state = 0

    def pause(self) -> None:
        """Pauses the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is not None:
            if self._player.get_state() != self.vlc.State.Opening:
                self._player.pause()
                self._state = 2

    def seek(self, direction, amount) -> None:
        """Seek forward or backward in the media.

        Overrides method from Player; see documentation in that class.
        """
        assert direction == 1 or direction == -1
        if self._player is not None:
            self._player.set_time(
                self._player.get_time() + (direction * amount * 1000)
            )

    @property
    def duration(self) -> int:
        """int: the duration of the player"""
        result = 0
        if self._media is not None:
            self._duration = self._media.get_duration()
            result = self._duration
        return result

    @property
    def time(self) -> int:
        """int: the current time of the player"""
        if self._player is not None:
            return self._player.get_time()

    @property
    def time_str(self) -> str:
        """str: the formatted time and duration of the player"""
        result = "00:00:00/00:00:00"
        if self._player is not None:
            time_seconds = int(self.time / 1000)
            length_seconds = int(self.duration / 1000)
            t = time.strftime('%H:%M:%S', time.gmtime(time_seconds))
            d = time.strftime('%H:%M:%S', time.gmtime(length_seconds))
            result = "%s/%s" % (t, d)
        return result



import os
import threading

from castero import helpers
from castero.config import Config
from castero.datafile import DataFile


class Episode:
    """The Episode class.

    This class represents a single episode from a podcast feed.
    """

    def __init__(self, feed, ep_id=None, title=None, description=None, link=None,
                 pubdate=None, copyright=None, enclosure=None, played=False) -> None:
        """Initializes the object.

        At least one of a title or description must be specified.

        Args:
            feed: the feed that this episode is a part of
            title: (optional) the title of the episode
            description: (optional) the description of the episode
            link: (optional) a link to the episode
            pubdate: (optional) the date the episode was published, as a string
            copyright: (optional) the copyright notice of the episode
            enclosure: (optional) a url to a media file
            played: (optional) whether the episode has been played
        """
        assert title is not None or description is not None

        self._feed = feed
        self._ep_id = ep_id
        self._title = title
        self._description = description
        self._link = link
        self._pubdate = pubdate
        self._copyright = copyright
        self._enclosure = enclosure
        self._played = played

    def __str__(self) -> str:
        """Represent this object as a single-line string.

        Returns:
            string: this episode's title, if it exists, else its description
        """
        if self._title is not None:
            representation = self._title
        else:
            representation = self._description
        return representation.split('\n')[0]

    def _feed_directory(self) -> str:
        """Gets the path to the downloaded episode's feed directory.

        This method does not ensure whether the directory exists -- it simply
        acts as a single definition of where it _should_ be.

        Returns:
            str: a path to the feed directory
        """
        feed_dirname = helpers.sanitize_path(str(self._feed))
        if Config is None or Config["custom_download_dir"] == "":
            path = DataFile.DEFAULT_DOWNLOADED_DIR
        else:
            path = Config["custom_download_dir"]
        return os.path.join(path, feed_dirname)

    def get_playable(self) -> str:
        """Gets a playable path for this episode.

        This method checks whether the episode is available on the disk, giving
        the path to that file if so. Otherwise, simply return the episode's
        enclosure, which is probably a URL.

        Returns:
            str: a path to a playable file for this episode
        """
        playable = self.enclosure

        episode_partial_filename = helpers.sanitize_path(str(self))
        feed_directory = self._feed_directory()

        if os.path.exists(feed_directory):
            for File in os.listdir(feed_directory):
                if File.startswith(episode_partial_filename + '.'):
                    playable = os.path.join(feed_directory, File)

        return playable

    def download(self, download_queue, display=None):
        """Downloads this episode to the file system.

        This method currently only supports downloading from an external URL.
        In the future, it may be worthwhile to determine whether the episode's
        source is a local file and simply copy it instead.

        Args:
            download_queue: the download_queue overseeing this download
            display: (optional) the display to write status updates to
        """
        if self._enclosure is None:
            if display is not None:
                display.change_status("Download failed: episode does not have"
                                      " a valid media source")
            return

        feed_directory = self._feed_directory()
        episode_partial_filename = helpers.sanitize_path(str(self))
        extension = os.path.splitext(self._enclosure)[1].split('?')[0]
        output_path = os.path.join(feed_directory,
                                   episode_partial_filename + str(extension))
        DataFile.ensure_path(output_path)

        if display is not None:
            display.change_status("Starting episode download...")

        t = threading.Thread(
            target=DataFile.download_to_file,
            args=[
                self._enclosure, output_path, str(self),
                download_queue, display
            ],
            name="download_%s" % str(self)
        )
        t.start()

    def delete(self, display=None):
        """Deletes the episode file from the file system.

        Args:
            display: (optional) the display to write status updates to
        """
        if self.downloaded:
            episode_partial_filename = helpers.sanitize_path(str(self))
            feed_directory = self._feed_directory()

            if os.path.exists(feed_directory):
                for File in os.listdir(feed_directory):
                    if File.startswith(episode_partial_filename + '.'):
                        os.remove(os.path.join(feed_directory, File))
                        if display is not None:
                            display.change_status(
                                "Successfully deleted the downloaded episode"
                            )

                # if there are no more files in the feed directory, delete it
                if len(os.listdir(feed_directory)) == 0:
                    os.rmdir(feed_directory)

    def downloaded(self) -> bool:
        """Determines whether the episode is downloaded.

        Returns:
            bool: whether or not the episode is downloaded
        """
        found_downloaded = False
        episode_partial_filename = helpers.sanitize_path(str(self))
        feed_directory = self._feed_directory()

        if os.path.exists(feed_directory):
            for File in os.listdir(feed_directory):
                if File.startswith(episode_partial_filename + '.'):
                    found_downloaded = True
        return found_downloaded

    @property
    def ep_id(self) -> int:
        """int: the database id of the episode"""
        return self._ep_id

    @ep_id.setter
    def ep_id(self, ep_id) -> None:
        self._ep_id = ep_id

    @property
    def title(self) -> str:
        """str: the title of the episode"""
        result = self._title
        if result is None:
            result = "Title not available."
        return result

    @property
    def description(self) -> str:
        """str: the description of the episode"""
        result = self._description
        if result is None:
            result = "Description not available."
        return result

    @property
    def link(self) -> str:
        """str: the link of/for the episode"""
        result = self._link
        if result is None:
            result = "Link not available."
        return result

    @property
    def pubdate(self) -> str:
        """str: the publish date of the episode"""
        result = self._pubdate
        if result is None:
            result = "Publish date not available."
        return result

    @property
    def copyright(self) -> str:
        """str: the copyright of the episode"""
        result = self._copyright
        if result is None:
            result = "No copyright specified."
        return result

    @property
    def enclosure(self) -> str:
        """str: the enclosure of the episode"""
        result = self._enclosure
        if result is None:
            result = "Enclosure not available."
        return result

    @property
    def played(self) -> bool:
        """bool: whether the episode has been played"""
        return self._played

    @played.setter
    def played(self, played) -> None:
        self._played = played

    @property
    def metadata(self) -> str:
        """str: the user-displayed metadata of the episode"""
        description = helpers.html_to_plain(self.description) if \
            helpers.is_true(Config["clean_html_descriptions"]) else \
            self.description
        description = description.replace('\n', '')
        downloaded = "Episode downloaded and available for offline playback." \
            if self.downloaded() else "Episode not downloaded."

        return \
            f"\cb{self.title}\n" \
            f"{self.pubdate}\n\n" \
            f"{self.link}\n\n" \
            f"\cbDescription:\n" \
            f"{description}\n\n" \
            f"\cbCopyright:\n" \
            f"{self.copyright}\n\n" \
            f"\cbDownloaded:\n" \
            f"{downloaded}\n"



import curses
import glob
import importlib
import threading
from typing import List
from os.path import dirname, basename, isfile

import castero
from castero import helpers
from castero.config import Config
from castero.database import Database
from castero.downloadqueue import DownloadQueue
from castero.feed import Feed, FeedError, FeedLoadError, FeedDownloadError, \
    FeedParseError, FeedStructureError
from castero.episode import Episode
from castero.perspective import Perspective
from castero.queue import Queue


class DisplayError(Exception):
    """An ambiguous error while handling the display.
    """


class DisplaySizeError(DisplayError):
    """The display does not have acceptable dimensions.
    """


class Display:
    """The Display class.

    This class is used to handle all user-interaction. It creates and handles
    all aspects of the application's interface, including windows and menus. It
    retrieves input from the user and performs corresponding actions.
    """
    MIN_WIDTH = 20
    MIN_HEIGHT = 8
    INPUT_TIMEOUT = 1000  # 1 second
    STATUS_TIMEOUT = 4  # multiple of INPUT_TIMEOUT
    AVAILABLE_COLORS = {
        'black': curses.COLOR_BLACK,
        'blue': curses.COLOR_BLUE,
        'cyan': curses.COLOR_CYAN,
        'green': curses.COLOR_GREEN,
        'magenta': curses.COLOR_MAGENTA,
        'red': curses.COLOR_RED,
        'white': curses.COLOR_WHITE,
        'yellow': curses.COLOR_YELLOW,
        'transparent': -1
    }
    KEY_MAPPING = {chr(i): i for i in range(256)}
    KEY_MAPPING.update(
        (name[4:], value) for name, value in vars(curses).items()
        if name.startswith('KEY_')
    )
    KEY_MAPPING.update(
        {
            'ENTER': 10,
            'SPACE': 32
        }
    )
    AVAILABLE_PLAYERS = {}

    def __init__(self, stdscr, database) -> None:
        """Initializes the object.

        Args:
            stdscr: a stdscr from curses.initscr()
            database: a connected castero.Database
        """
        self._stdscr = stdscr
        self._database = database
        self._parent_x = -1
        self._parent_y = -1
        self._perspectives = {}
        self._active_perspective = 1
        self._header_window = None
        self._footer_window = None
        self._queue = Queue(self)
        self._download_queue = DownloadQueue(self)
        self._status = ""
        self._status_timer = self.STATUS_TIMEOUT
        self._menus_valid = True
        self._modified_episodes = []

        # basic preliminary operations
        self._stdscr.timeout(self.INPUT_TIMEOUT)
        curses.start_color()
        curses.noecho()
        curses.curs_set(0)
        curses.cbreak()
        self._stdscr.keypad(True)

        self.update_parent_dimensions()
        self.create_color_pairs()
        self._load_perspectives()
        self._load_players()
        self._create_windows()
        self.create_menus()

    def create_color_pairs(self) -> None:
        """Initializes color pairs used for the display.

        Creates the following color pairs (foreground, background):
            - 1: foreground, background
            - 2: background, foreground
            - 3: background_alt, foreground_alt
            - 4: foreground_alt, background_alt
        """
        assert Config["color_foreground"] in self.AVAILABLE_COLORS
        assert Config["color_background"] in self.AVAILABLE_COLORS
        assert Config["color_foreground_alt"] in self.AVAILABLE_COLORS
        assert Config["color_background_alt"] in self.AVAILABLE_COLORS

        if (self.AVAILABLE_COLORS[Config["color_background"]] == -1 or
                self.AVAILABLE_COLORS[Config["color_background_alt"]] == -1):
            curses.use_default_colors()

        curses.init_pair(
            1,
            self.AVAILABLE_COLORS[Config["color_foreground"]],
            self.AVAILABLE_COLORS[Config["color_background"]]
        )
        curses.init_pair(
            2,
            self.AVAILABLE_COLORS[Config["color_background"]],
            self.AVAILABLE_COLORS[Config["color_foreground"]]
        )
        curses.init_pair(
            3,
            self.AVAILABLE_COLORS[Config["color_background_alt"]],
            self.AVAILABLE_COLORS[Config["color_foreground_alt"]]
        )
        curses.init_pair(
            4,
            self.AVAILABLE_COLORS[Config["color_foreground_alt"]],
            self.AVAILABLE_COLORS[Config["color_background_alt"]]
        )
        curses.init_pair(
            5,
            self.AVAILABLE_COLORS[Config["color_foreground_dim"]],
            self.AVAILABLE_COLORS[Config["color_background"]]
        )

    def _load_perspectives(self) -> None:
        """Load instances of perspectives from the `perspectives` package.
        """
        # load a list of modules names by manually detecting .py files
        module_files = glob.glob(dirname(__file__) + "/perspectives/*.py")
        module_names = [basename(f)[:-3] for f in module_files if isfile(f)]

        for name in module_names:
            p_mod = importlib.import_module("castero.perspectives.%s" % name)
            p_cls = getattr(
                p_mod,
                dir(p_mod)[[cls.lower() == name
                            for cls in dir(p_mod)].index(True)])
            inst = p_cls(self)
            self._perspectives[inst.ID] = inst

    def _load_players(self) -> None:
        """Load player classes from the `players` package.
        """
        # load a list of modules names by manually detecting .py files
        module_files = glob.glob(dirname(__file__) + "/players/*.py")
        module_names = [basename(f)[:-3] for f in module_files if isfile(f)]

        for name in module_names:
            p_mod = importlib.import_module("castero.players.%s" % name)
            p_cls = getattr(
                p_mod,
                dir(p_mod)[[cls.lower() == name
                            for cls in dir(p_mod)].index(True)])
            self.AVAILABLE_PLAYERS[p_cls.NAME] = p_cls

    def _create_windows(self) -> None:
        """Creates and sets basic parameters for the windows.

        If the windows already exist when this method is run, this method will
        delete them and create new ones.
        """
        # delete old windows if they exist
        if self._header_window is not None:
            del self._header_window
            self._header_window = None
        if self._footer_window is not None:
            del self._footer_window
            self._footer_window = None

        # create windows
        self._header_window = curses.newwin(2, self._parent_x,
                                            0, 0)
        self._footer_window = curses.newwin(2, self._parent_x,
                                            self._parent_y - 2, 0)

        # set window attributes
        self._header_window.attron(curses.color_pair(4))
        self._footer_window.attron(curses.color_pair(4))

        # create windows for all perspectives
        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].create_windows()

    def create_menus(self) -> None:
        """Creates the menus used in each window.

        Windows which have menus should be created prior to running this method
        (using _create_windows).
        """
        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].create_menus()

    def show_help(self) -> None:
        """Show the help screen.

        This method takes over the main loop, displaying the screen until a key
        is pressed. This means that typical loop actions, including checking
        the state of the current player, will not run while this screen is up.
        """
        self.clear()
        self._stdscr.refresh()

        padding_yx = (1, 4)

        help_window = curses.newwin(self._parent_y, self._parent_x, 0, 0)
        help_window.attron(curses.A_BOLD)

        # display lines from __help__
        help_lines = \
            castero.__help__.split('\n')[:self._parent_y - padding_yx[0] - 1]
        help_lines.append("Press any key to exit this screen.")
        for i in range(len(help_lines)):
            help_window.addstr(i + padding_yx[0], padding_yx[1],
                               help_lines[i][:self._parent_x - padding_yx[1]])
        help_window.refresh()

        # simply wait until any key is pressed (temporarily disable timeout)
        self._stdscr.timeout(-1)
        self._stdscr.getch()
        self._stdscr.timeout(self.INPUT_TIMEOUT)

        self.clear()

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders.
        """
        # check if the screen size has changed
        self.update_parent_dimensions()

        # check to see if menu contents have been invalidated
        if not self.menus_valid:
            for perspective_id in self._perspectives:
                self._perspectives[perspective_id].update_menus()
            self.menus_valid = True

        # add header
        playing_str = castero.__title__
        if self._queue.first is not None:
            state = self._queue.first.state
            playing_str = ["Stopped", "Playing", "Paused"][state] + \
                          ": %s" % self._queue.first.title
            if self._queue.length > 1:
                playing_str += " (+%d in queue)" % (self._queue.length - 1)

            if helpers.is_true(Config["right_align_time"]):
                playing_str += ("[%s]" % self._queue.first.time_str).rjust(
                    self._header_window.getmaxyx()[1] - len(playing_str))
            else:
                playing_str += " [%s]" % self._queue.first.time_str

        self._header_window.attron(curses.A_BOLD)
        self._header_window.addstr(0, 0, " " * self._parent_x)
        self._header_window.addstr(0, 0, playing_str)

        # add footer
        footer_str = ""
        if self._status == "" and not \
                helpers.is_true(Config["disable_default_status"]):
            feeds = self.database.feeds()
            if len(feeds) > 0:
                total_feeds = len(feeds)
                lengths_of_feeds = \
                    [len(self.database.episodes(feed)) for feed in feeds]
                total_episodes = sum(lengths_of_feeds)
                median_episodes = helpers.median(lengths_of_feeds)

                footer_str += "Found %d feeds with %d total episodes (avg." \
                              " %d episodes, med. %d)" % (
                                  total_feeds,
                                  total_episodes,
                                  total_episodes / total_feeds,
                                  median_episodes
                              )
            else:
                footer_str += "No feeds added"
        else:
            footer_str = self._status

        if footer_str != "":
            footer_str += " -- Press %s for help" % Config["key_help"]

        self._footer_window.attron(curses.A_BOLD)
        self._footer_window.addstr(
            1, 0, " " * (self._footer_window.getmaxyx()[1] - 1)
        )
        footer_str = footer_str[:self._footer_window.getmaxyx()[1] - 1]
        self._footer_window.addstr(1, 0, footer_str)

        # add window borders
        self._header_window.hline(1, 0,
                                  0, self._header_window.getmaxyx()[1])
        self._footer_window.hline(0, 0,
                                  0, self._footer_window.getmaxyx()[1])

        # update display for current perspective
        self._perspectives[self._active_perspective].display()

    def _get_active_perspective(self) -> Perspective:
        """Retrieve the active/visible Perspective.
        """
        return self._perspectives[self._active_perspective]

    def _change_active_perspective(self, perspective_id) -> None:
        """Changes _active_perspective to the given perspective.

        Args:
            perspective_id: the ID of the perspective to change to
        """
        assert perspective_id in self._perspectives

        self._active_perspective = perspective_id
        self._perspectives[perspective_id].made_active()
        self.clear()

    def _get_input_str(self, prompt) -> str:
        """Prompts the user for input and returns the resulting string.

        This method assumes that all input strings will be obtained in the
        footer window.

        Args:
            prompt: a string to inform the user of what they need to enter

        Returns:
            str: the user's input
        """
        assert self._footer_window is not None
        assert isinstance(prompt, str)

        curses.curs_set(1)
        self._stdscr.timeout(-1)  # disable timeouts while waiting for entry

        # display input prompt
        self._footer_window.addstr(
            1, 0, " " * (self._footer_window.getmaxyx()[1] - 1)
        )
        self._footer_window.addstr(1, 0, prompt)

        entry_pad = curses.newpad(1, 999)
        current_x = 0
        scroll_x = 0
        input_char = None
        while input_char not in [curses.KEY_ENTER, 10]:
            if input_char is not None:
                # manually handle backspace
                if input_char in [curses.KEY_BACKSPACE, 127]:
                    if current_x > 0:
                        entry_pad.delch(0, current_x - 1)
                        current_x -= 1
                        if scroll_x > 0:
                            scroll_x -= 1
                else:
                    # scroll the input pad if necessary
                    if current_x + len(prompt) > \
                            self._footer_window.getmaxyx()[1] - 1:
                        scroll_x += 1

                    # add the entered character to the pad
                    entry_pad.addch(0, current_x, input_char)
                    current_x += 1

                # display current portion of pad
                entry_pad.refresh(0, scroll_x,
                                  self._parent_y - 1, len(prompt),
                                  self._parent_y - 1,
                                  self._footer_window.getmaxyx()[1] - 1)

            # get the next input character
            input_char = self._footer_window.getch()

        self._stdscr.timeout(self.INPUT_TIMEOUT)
        self._footer_window.clear()
        curses.curs_set(0)

        return entry_pad.instr(0, 0, entry_pad.getmaxyx()[1]) \
            .decode('utf-8').strip()

    def _get_y_n(self, prompt) -> bool:
        """Prompts the user for a yes or no (y/n) input.

        Args:
            prompt: a string to inform the user of what they need to enter

        Returns:
            bool: true if the user presses y, false otherwise
        """
        assert self._footer_window is not None
        assert isinstance(prompt, str)

        curses.echo()
        curses.curs_set(1)

        self._footer_window.addstr(
            1, 0, " " * (self._footer_window.getmaxyx()[1] - 1)
        )
        self._footer_window.addstr(1, 0, prompt)
        char = self._footer_window.getch()

        self._footer_window.clear()
        curses.curs_set(0)
        curses.noecho()

        return char == ord('y')

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        Args:
            c: the input character

        Returns:
            bool: whether or not the application should continue running
        """
        for perspective_id in self._perspectives:
            if c == self.KEY_MAPPING[str(perspective_id)]:
                self._change_active_perspective(perspective_id)

        return self._get_active_perspective().handle_input(c)

    def add_feed(self) -> None:
        """Prompt the user for a feed and add it, if possible.
        """
        path = self._get_input_str("Enter the URL or path of the feed: ")
        try:
            # assume urls have http in them
            if "http" in path:
                feed = Feed(url=path)
            else:
                feed = Feed(file=path)
            if feed.validated:
                self.database.replace_feed(feed)
                self.database.replace_episodes(feed, feed.parse_episodes())
            self.menus_valid = False
            self.change_status("Feed '%s\' successfully added" % str(feed))
        except FeedError as e:
            if isinstance(e, FeedLoadError):
                self.change_status(
                    "FeedLoadError: %s" % str(e)
                )
            elif isinstance(e, FeedDownloadError):
                self.change_status(
                    "FeedDownloadError: %s" % str(e)
                )
            elif isinstance(e, FeedParseError):
                self.change_status(
                    "FeedParseError: %s" % str(e)
                )
            elif isinstance(e, FeedStructureError):
                self.change_status(
                    "FeedStructureError: %s" % str(e)
                )
            else:
                self.change_status(
                    "FeedError [ambiguous]: %s" % str(e)
                )

    def delete_feed(self, feed: Feed) -> None:
        """Deletes the given feed from the database.

        If the delete_feed_confirmation config option is true, this method will
        first ask for y/n confirmation before deleting the feed.

        Deleting a feed also deletes all downloaded/saved episodes.

        Args:
            feed: the Feed to delete, which can be None
        """
        if feed is not None:
            should_delete = True
            if helpers.is_true(Config["delete_feed_confirmation"]):
                should_delete = self._get_y_n(
                    "Are you sure you want to delete this feed? (y/n): "
                )
            if should_delete:
                self.database.delete_feed(feed)
                self.menus_valid = False
                self.change_status("Feed successfully deleted")

    def reload_feeds(self) -> None:
        """Reloads the users' feeds.

        If the total number of feeds is >= the reload_feeds_threshold config
        option, this method will first ask for y/n confirmation.

        This method starts the reloading in a new un-managed thread.
        """
        should_reload = True
        if len(self.database.feeds()) >= int(Config["reload_feeds_threshold"]):
            should_reload = self._get_y_n(
                "Are you sure you want to reload all of your feeds?"
                " (y/n): "
            )
        if should_reload:
            t = threading.Thread(target=self.database.reload, args=[self])
            t.start()

    def save_episodes(self, feed=None, episode=None) -> None:
        """Save a feed or episode.

        If the user is saving an episode and the episode is already saved, this
        method will instead ask the user if they would like to delete the
        downloaded episode. However, if the user is saving a feed, there is no
        prompt to delete episodes, even if some are downloaded. In this case,
        downloaded episodes are simply skipped.

        Exactly one of either feed or episode must be given.

        Args:
            feed: (optional) a feed to download all episodes of
            episode: (optional) an episode to download or delete
        """
        assert (feed is None or episode is None) and (feed is not episode)

        if feed is not None:
            for episode in self.database.episodes(feed):
                if not episode.downloaded():
                    self._download_queue.add(episode)
        else:
            if episode.downloaded():
                should_delete = self._get_y_n(
                    "Are you sure you want to delete the downloaded"
                    " episode? (y/n): ")
                if should_delete:
                    episode.delete(self)
            else:
                self._download_queue.add(episode)

    def clear(self) -> None:
        """Clear the screen.
        """
        self._stdscr.clear()

    def refresh(self) -> None:
        """Refresh the screen and all windows in all perspectives.
        """
        self._stdscr.refresh()

        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].refresh()

        self._header_window.refresh()
        self._footer_window.refresh()

    def terminate(self) -> None:
        """Set console settings to their normal state.

        This method does not, by itself, cause the application to exit. Nor
        does it even cause the input loop to end. It should simply be seen as
        a "wrapping up" method for any actions which need to be performed
        before the object is destroyed.
        """
        self._queue.stop()

        curses.nocbreak()
        self._stdscr.keypad(False)
        curses.echo()
        curses.endwin()

    def update_parent_dimensions(self) -> None:
        """Update _parent_x and _parent_y to the size of the console.
        """
        current_y, current_x = self._stdscr.getmaxyx()

        if current_y != self._parent_y or current_x != self._parent_x:
            self._parent_y, self._parent_x = current_y, current_x
            self._create_windows()
            self.menus_valid = False
            self.refresh()

        if self._parent_y < self.MIN_HEIGHT:
            raise DisplaySizeError("Display height is too small")
        if self._parent_x < self.MIN_WIDTH:
            raise DisplaySizeError("Display width is too small")

    def getch(self) -> int:
        """Gets an input character from the user.

        This method returns after at most INPUT_TIMEOUT ms.

        Returns:
            int: the character entered by the user, or -1
        """
        char = self._stdscr.getch()
        return char

    def change_status(self, status) -> None:
        """Changes the status message displayed in the footer.

        Args:
            status: the status message to display
        """
        assert isinstance(status, str)

        self._status = status
        self._status_timer = self.STATUS_TIMEOUT

    def update(self) -> None:
        """Updates all actively tracked components of this object.

        Should be called by the main loop after every input or input timeout.
        """
        # have the queue check if it needs to go to the next player
        self._queue.update()

        # check the status of any downloads
        try:
            self._download_queue.update()
        except OSError as e:
            self.change_status("OSError: %s" % str(e))
            return

        # update the status timer
        # If the user is not doing anything, the status message will take
        # INPUT_TIMEOUT * STATUS_TIMEOUT ms to be cleared. However, if the user
        # is performing inputs (i.e. traversing a menu) the message may be
        # cleared much quicker, since it will go away in STATUS_TIMEOUT
        # keypresses. However, this seems reasonable, since if the user is
        # actively controlling the client and not pausing to read the message,
        # they probably don't care about it anyway.
        if self._status_timer > 0:
            self._status_timer -= 1
            if self._status_timer <= 0:
                # status_timer should be reset during the next change_status()
                self._status = ""

        # write any episode modifications to the database
        if len(self._modified_episodes) > 0:
            for episode in self._modified_episodes:
                self.database.replace_episode(episode._feed, episode)
            self.menus_valid = False
            self._modified_episodes = []

    @property
    def parent_x(self) -> int:
        """int: the width of the parent screen, in characters"""
        return self._parent_x

    @property
    def parent_y(self) -> int:
        """int: the height of the parent screen, in characters"""
        return self._parent_y

    @property
    def database(self) -> Database:
        """Database: the user's database"""
        return self._database

    @property
    def perspectives(self) -> dict:
        """dict: the loaded Perspective's with id:perspective pairs"""
        return self._perspectives

    @property
    def queue(self) -> Queue:
        """Queue: the Queue of Player's"""
        return self._queue

    @property
    def menus_valid(self) -> bool:
        """bool: whether the menu contents are valid (!need_to_be_updated)"""
        return self._menus_valid

    @menus_valid.setter
    def menus_valid(self, menus_valid) -> None:
        self._menus_valid = menus_valid

    @property
    def modified_episodes(self) -> List[Episode]:
        """List[Episode]: database episodes to save on the next update"""
        return self._modified_episodes


from castero.config import Config
from castero.player import Player


class Queue:
    """The Queue class.

    A FIFO ordered queue of Player instances. This class is also the display
    class' main interface for accessing information about the current player.
    """

    def __init__(self, display) -> None:
        """Initializes the object.
        """
        self._players = []
        self._display = display

    def __getitem__(self, index):
        return self._players[index]

    def __iter__(self):
        return self._players.__iter__()

    def clear(self) -> None:
        """Clears the queue.

        Tt is extremely likely that the caller of this method will also want to
        stop the first player prior to calling this method.
        """
        for p in self._players:
            del p
        self._players = []

    def next(self) -> None:
        """Proceed to the next player in the queue.
        """
        if len(self._players) > 0:
            self._players.pop(0)
            self.play()

    def add(self, player) -> None:
        """Adds a player to the end of the queue.
        """
        assert isinstance(player, Player)

        self._players.append(player)

    def play(self) -> None:
        """Plays the first player in the queue.
        """
        if self.first is not None:
            self.first.episode.played = True
            self._display.modified_episodes.append(self.first.episode)
            self.first.play()

    def pause(self) -> None:
        """Pauses the first player in the queue.
        """
        if self.first is not None:
            self.first.pause()

    def stop(self) -> None:
        """Stops the first player in the queue."""
        if self.first is not None:
            self.first.stop()

    def toggle(self) -> None:
        """Plays or pauses the first player.
        """
        if self.first is not None:
            if self.first.state == 1:  # playing
                self.pause()
            else:  # paused or stopped
                self.play()

    def seek(self, direction) -> None:
        """Seeks the first player in the specified direction.

        Args:
            direction: 1 to move forward, -1 to move backward
        """
        assert direction == 1 or direction == -1

        if self.first is not None:
            distance = int(Config["seek_distance_forward" if direction == 1 else "seek_distance_backward"])
            self.first.seek(direction, distance)

    def remove(self, player) -> int:
        """Remove a player from the queue, if it is currently in it.

        Args:
            player: the Player to remove

        Returns:
            int: the index of the player in the queue, or -1
        """
        result = -1
        if player in self._players:
            result = self._players.index(player)
            self._players.remove(player)
        return result

    def update(self) -> None:
        """Checks the status of the current player.
        """
        if self.first is not None:
            # sanity check the player's current time
            if self.first.duration > 0:
                if (self.first.time / 1000) + 1 >= \
                        (self.first.duration / 1000):
                    self.next()

    @property
    def first(self) -> Player:
        """Player: the first player in the queue"""
        result = None
        if len(self._players) > 0:
            result = self._players[0]
        return result

    @property
    def length(self) -> int:
        """int: the length of the queue"""
        return len(self._players)



import curses
from abc import ABC, abstractmethod

import cjkwrap

from castero import helpers
from castero.config import Config
from castero.menu import Menu


class Perspective(ABC):
    """The Perspective class.

    This class is extended by perspectives -- classes which offer methods to
    handle display elements with a certain layout. Perspectives only control
    the "inside" elements of the display: the header and footer are controlled
    by the display class.

    The user is able to switch between perspectives by using the 0-9 keys
    corresponding to the perspective's ID.

    Instances of this class do not generally hold data variables, i.e. the
    instance of the Feed class or Config class. We instead reference the
    variables held in Display instance.
    """
    ID = -1

    @abstractmethod
    def __init__(self, display):
        """Initializes the object.

        This method does not automatically create configure some of its
        necessary elements, i.e. the windows. Instead, for some methods it is
        expected that the Display instance will call identically-named methods
        for all of its "children" perspectives when "chain" methods are called.

        For example, the Display class has a _create_windows method, which
        should call this class's create_windows method.

        Args:
            display: the parent Display instance
        """
        self._display = display

    @abstractmethod
    def create_windows(self) -> None:
        """Create and set basic parameters for the windows.
        """

    @abstractmethod
    def create_menus(self) -> None:
        """Create the menus used in each window.

        Windows which have menus should be created prior to running this method
        (using _create_windows).
        """

    @abstractmethod
    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders.
        """

    @abstractmethod
    def refresh(self) -> None:
        """Refresh the screen and all windows.
        """

    @abstractmethod
    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        Args:
            c: the input character

        Returns:
            bool: whether or not the application should continue running
        """

    @abstractmethod
    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to).
        """

    @abstractmethod
    def update_menus(self) -> None:
        """Update/refresh the contents of all menus.
        """

    @abstractmethod
    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one.

        Returns:
            Menu: the active Menu, or None
        """

    def _change_active_window(self, direction) -> None:
        """Changes _active_window to the next or previous window, if available.

        Args:
            direction: 1 to change to the next window, -1 to change to the
                previous (if it exists)
        """
        assert direction == 1 or direction == -1

        active_menu = self._get_active_menu()
        if active_menu is not None:
            active_menu.set_active(False)

        self._active_window += direction
        if self._active_window > 1:
            self._active_window = 1
        elif self._active_window < 0:
            self._active_window = 0

        new_active_menu = self._get_active_menu()
        if new_active_menu is not None:
            new_active_menu.set_active(True)

    def _draw_metadata(self, window) -> None:
        """Draws the metadata of the selected feed/episode onto the window.

        Exactly one of feed or episode must be specified.

        Args:
            window: the curses window which will display the metadata
        """
        assert window is not None

        output_lines = []

        max_lines = int(0.7 * window.getmaxyx()[0])
        max_line_width = window.getmaxyx()[1] - 1

        # clear the window by drawing blank lines
        for y in range(2, window.getmaxyx()[0]):
            window.addstr(y, 0, " " * max_line_width)

        metadata = self._get_active_menu().metadata()
        temp_lines = metadata.split('\n')

        lines = []
        for line in temp_lines:
            parts = cjkwrap.wrap(line, window.getmaxyx()[1] - 1,
                                 replace_whitespace=True)
            if len(parts) == 0:
                lines.append('')
            else:
                for part in parts:
                    lines.append(part.strip())

        # truncate to at most 70% of the total lines on the screen
        lines = lines[:max_lines]

        y = 2
        for line in lines[:max_lines]:
            if line.startswith('\cb'):
                window.attron(curses.A_BOLD)
                line = line[3:]
            else:
                window.attrset(curses.color_pair(1))

            window.addstr(y, 0, line)
            y += 1



import curses

from castero import helpers
from castero.config import Config
from castero.menu import Menu
from castero.menus.episodemenu import EpisodeMenu
from castero.menus.feedmenu import FeedMenu
from castero.perspective import Perspective
from castero.player import Player


class PrimaryPerspective(Perspective):
    """The PrimaryPerspective class.

    This class handles display elements while in the primary perspective, which
    is the default perspective.
    """
    ID = 1

    def __init__(self, display) -> None:
        """Initializes the object.

        Overrides method from Perspective; see documentation in that class.
        """
        super().__init__(display)
        self._active_window = 0
        self._feed_window = None
        self._episode_window = None
        self._metadata_window = None
        self._feed_menu = None
        self._episode_menu = None
        self._metadata_updated = False

    def create_windows(self) -> None:
        """Create and set basic parameters for the windows.

        Overrides method from Perspective; see documentation in that class.
        """
        # delete old windows if they exist
        if self._feed_window is not None:
            del self._feed_window
            self._feed_window = None
        if self._episode_window is not None:
            del self._episode_window
            self._episode_window = None
        if self._metadata_window is not None:
            del self._metadata_window
            self._metadata_window = None

        parent_x = self._display.parent_x
        parent_y = self._display.parent_y
        third_x = helpers.third(parent_x)
        self._feed_window = curses.newwin(parent_y - 2, third_x,
                                          2, 0)
        self._episode_window = curses.newwin(parent_y - 2, third_x,
                                             2, third_x)
        metadata_width = parent_x - ((third_x * 2) - 1)
        self._metadata_window = curses.newwin(parent_y - 2, metadata_width,
                                              2, 2 * third_x)

        self._feed_window.attron(curses.color_pair(1))
        self._episode_window.attron(curses.color_pair(1))
        self._metadata_window.attron(curses.color_pair(1))

    def create_menus(self) -> None:
        """Create the menus used in each window.

        Overrides method from Perspective; see documentation in that class.
        """
        assert all(window is not None for window in [
            self._feed_window, self._episode_window
        ])

        self._episode_menu = EpisodeMenu(
            self._episode_window, self._display.database)
        self._feed_menu = FeedMenu(self._feed_window, self._display.database,
                                   child=self._episode_menu, active=True)

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders.

        Overrides method from Perspective; see documentation in that class.
        """
        # add window titles
        self._feed_window.attron(curses.A_BOLD)
        self._episode_window.attron(curses.A_BOLD)
        self._metadata_window.attron(curses.A_BOLD)
        self._feed_window.addstr(0, 0, "Feeds")
        self._episode_window.addstr(0, 0, "Episodes")
        self._metadata_window.addstr(0, 0, "Metadata")

        # add window borders
        self._feed_window.hline(1, 0,
                                0, self._feed_window.getmaxyx()[1])

        self._episode_window.hline(1, 0,
                                   0, self._episode_window.getmaxyx()[1])
        self._metadata_window.hline(1, 0,
                                    0, self._metadata_window.getmaxyx()[1] - 1)
        if not helpers.is_true(Config["disable_vertical_borders"]):
            self._feed_window.vline(0, self._feed_window.getmaxyx()[1] - 1,
                                    0, self._feed_window.getmaxyx()[0] - 2)
            self._episode_window.vline(0,
                                       self._episode_window.getmaxyx()[1] - 1,
                                       0,
                                       self._episode_window.getmaxyx()[0] - 2)

        # display menu content
        self._feed_menu.display()
        self._episode_menu.display()

        # draw metadata
        if not self._metadata_updated:
            self._draw_metadata(self._metadata_window)

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        Overrides method from Perspective; see documentation in that class.
        """
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config['key_exit']]:
            self._display.terminate()
            keep_running = False
        elif c == key_mapping[Config['key_help']]:
            self._display.show_help()
        elif c == key_mapping[Config['key_right']]:
            self._change_active_window(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_left']]:
            self._change_active_window(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_up']]:
            self._get_active_menu().move(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_down']]:
            self._get_active_menu().move(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_scroll_up']]:
            self._get_active_menu().move_page(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_scroll_down']]:
            self._get_active_menu().move_page(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_play_selected']]:
            queue.stop()
            queue.clear()
            self._create_player_from_selected()
            queue.play()
        elif c == key_mapping[Config['key_add_selected']]:
            self._create_player_from_selected()
            self._get_active_menu().move(-1)
        elif c == key_mapping[Config['key_clear']]:
            queue.stop()
            queue.clear()
        elif c == key_mapping[Config['key_pause_play']] or \
                c == key_mapping[Config['key_pause_play_alt']]:
            queue.toggle()
        elif c == key_mapping[Config['key_next']]:
            queue.stop()
            queue.next()
            queue.play()
        elif c == key_mapping[Config['key_seek_forward']] or \
                c == key_mapping[Config['key_seek_forward_alt']]:
            queue.seek(1)
        elif c == key_mapping[Config['key_seek_backward']] or \
                c == key_mapping[Config['key_seek_backward_alt']]:
            queue.seek(-1)
        elif c == key_mapping[Config['key_add_feed']]:
            self._display.add_feed()
        elif c == key_mapping[Config['key_delete']]:
            if self._active_window == 0:
                self._display.delete_feed(self._feed_menu.item())
        elif c == key_mapping[Config['key_reload']]:
            self._display.reload_feeds()
        elif c == key_mapping[Config['key_save']]:
            if self._active_window == 0 and self._feed_menu.item():
                self._display.save_episodes(feed=self._feed_menu.item())
            elif self._active_window == 1 and self._episode_menu.item():
                self._display.save_episodes(episode=self._episode_menu.item())
        elif c == key_mapping[Config['key_invert']]:
            self._invert_selected_menu()
        elif c == key_mapping[Config['key_mark_played']]:
            if self._active_window == 0:
                feed = self._feed_menu.item()
                if feed is not None:
                    episodes = self._display.database.episodes(feed)
                    for episode in episodes:
                        episode.played = not episode.played
                    self._display.modified_episodes.extend(episodes)
            elif self._active_window == 1:
                episode = self._episode_menu.item()
                if episode is not None:
                    episode.played = not episode.played
                    self._display.modified_episodes.append(episode)
                    self._episode_menu.move(-1)

        return keep_running

    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to).

        Overrides method from Perspective; see documentation in that class.
        """

    def update_menus(self) -> None:
        """Update/refresh the contents of all menus.

        Overrides method from Perspective; see documentation in that class.
        """
        self._feed_menu.update_items(None)
        self._feed_menu.update_child()

    def refresh(self) -> None:
        """Refresh the screen and all windows.

        Overrides method from Perspective; see documentation in that class.
        """
        self._feed_window.refresh()
        self._episode_window.refresh()
        self._metadata_window.refresh()

    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one.

        Overrides method from Perspective; see documentation in that class.
        """
        assert 0 <= self._active_window < 2

        return {
            0: self._feed_menu,
            1: self._episode_menu,
        }.get(self._active_window)

    def _create_player_from_selected(self) -> None:
        """Creates player(s) based on the selected items and adds to the queue.

        If the active menu is the feed menu, then this will create players for
        all episodes in the selected feed. If the active menu is the episode
        menu, this will simply create a single player.

        This method will not clear the queue prior to adding the new player(s),
        nor will it play the episodes after running.
        """
        if self._active_window == 0:
            feed = self._feed_menu.item()
            if feed is not None:
                for episode in self._display.database.episodes(feed):
                    player = Player.create_instance(
                        self._display.AVAILABLE_PLAYERS, str(episode),
                        episode.get_playable(), episode)
                    self._display.queue.add(player)
        elif self._active_window == 1:
            episode = self._episode_menu.item()
            if episode is not None:
                player = Player.create_instance(
                    self._display.AVAILABLE_PLAYERS, str(episode),
                    episode.get_playable(), episode)
                self._display.queue.add(player)

    def _invert_selected_menu(self) -> None:
        """Inverts the contents of the selected menu.
        """
        self._get_active_menu().invert()
        if self._feed_menu:
            self._feed_menu.update_child()


import curses

from castero import helpers
from castero.config import Config
from castero.menu import Menu
from castero.menus.queuemenu import QueueMenu
from castero.perspective import Perspective


class QueuePerspective(Perspective):
    """The QueuePerspective class.

    This class handles display elements while in the queue perspective, which
    is a listing of the user's current Queue in which they can directly modify
    its state.
    """
    ID = 2

    def __init__(self, display) -> None:
        """Initializes the object.

        Overrides method from Perspective; see documentation in that class.
        """
        super().__init__(display)
        self._active_window = 0
        self._queue_window = None
        self._metadata_window = None
        self._queue_menu = None
        self._metadata_updated = False

    def create_windows(self) -> None:
        """Create and set basic parameters for the windows.

        Overrides method from Perspective; see documentation in that class.
        """
        # delete old windows if they exist
        if self._queue_window is not None:
            del self._queue_window
            self._queue_window = None
        if self._metadata_window is not None:
            del self._metadata_window
            self._metadata_window = None

        parent_x = self._display.parent_x
        parent_y = self._display.parent_y
        third_x = helpers.third(parent_x)
        self._queue_window = curses.newwin(parent_y - 2, third_x * 2,
                                           2, 0)
        metadata_width = parent_x - ((third_x * 2) - 1)
        self._metadata_window = curses.newwin(parent_y - 2, metadata_width,
                                              2, 2 * third_x)

        self._queue_window.attron(curses.color_pair(1))
        self._metadata_window.attron(curses.color_pair(1))

    def create_menus(self) -> None:
        """Create the menus used in each window.

        Overrides method from Perspective; see documentation in that class.
        """
        assert self._queue_window is not None

        self._queue_menu = QueueMenu(self._queue_window, self._display.queue,
                                     active=True)

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders.

        Overrides method from Perspective; see documentation in that class.
        """
        # add window titles
        self._queue_window.attron(curses.A_BOLD)
        self._metadata_window.attron(curses.A_BOLD)
        self._queue_window.addstr(0, 0, "Queue")
        self._metadata_window.addstr(0, 0, "Metadata")

        # add window borders
        self._queue_window.hline(1, 0,
                                 0, self._queue_window.getmaxyx()[1])
        self._metadata_window.hline(1, 0,
                                    0, self._metadata_window.getmaxyx()[1] - 1)
        if not helpers.is_true(Config["disable_vertical_borders"]):
            self._queue_window.vline(0, self._queue_window.getmaxyx()[1] - 1,
                                     0, self._queue_window.getmaxyx()[0] - 2)

        # display menu content
        self._queue_menu.display()

        # draw metadata
        if not self._metadata_updated:
            self._draw_metadata(self._metadata_window)

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        Overrides method from Perspective; see documentation in that class.
        """
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config['key_exit']]:
            self._display.terminate()
            keep_running = False
        elif c == key_mapping[Config['key_help']]:
            self._display.show_help()
        elif c == key_mapping[Config['key_right']]:
            self._change_active_window(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_left']]:
            self._change_active_window(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_up']]:
            self._get_active_menu().move(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_down']]:
            self._get_active_menu().move(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_scroll_up']]:
            self._get_active_menu().move_page(1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_scroll_down']]:
            self._get_active_menu().move_page(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config['key_play_selected']]:
            queue.stop()
            self._cycle_queue_to_selected()
            queue.play()
            self._display.menus_valid = False
        elif c == key_mapping[Config['key_pause_play']] or \
                c == key_mapping[Config['key_pause_play_alt']]:
            queue.toggle()
        elif c == key_mapping[Config['key_next']]:
            queue.stop()
            queue.next()
            queue.play()
            self._display.menus_valid = False
        elif c == key_mapping[Config['key_seek_forward']] or \
                c == key_mapping[Config['key_seek_forward_alt']]:
            queue.seek(1)
        elif c == key_mapping[Config['key_seek_backward']] or \
                c == key_mapping[Config['key_seek_backward_alt']]:
            queue.seek(-1)
        elif c == key_mapping[Config['key_clear']]:
            queue.stop()
            queue.clear()
            self._display.menus_valid = False
        elif c == key_mapping[Config['key_delete']]:
            self._remove_selected_from_queue()

        return keep_running

    def refresh(self) -> None:
        """Refresh the screen and all windows.

        Overrides method from Perspective; see documentation in that class.
        """
        self._queue_window.refresh()
        self._metadata_window.refresh()

    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to).

        Overrides method from Perspective; see documentation in that class.
        """
        pass

    def update_menus(self) -> None:
        """Update/refresh the contents of all menus.

        Overrides method from Perspective; see documentation in that class.
        """
        self._queue_menu.update_items(None)

    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one.

        Overrides method from Perspective; see documentation in that class.
        """
        assert 0 <= self._active_window < 2

        return self._queue_menu

    def _cycle_queue_to_selected(self) -> None:
        """Remove all players in the queue preceding the selected one.
        """
        target = self._queue_menu.item()
        while self._display.queue.first != target:
            self._display.queue.next()

    def _remove_selected_from_queue(self) -> None:
        """Remove the selected player from the queue.
        """
        player = self._queue_menu.item()
        if player is not None:
            index = self._display.queue.remove(player)
            self._queue_menu.update_items(None)
            for i in range(index):
                self._get_active_menu().move(-1)



import xml.etree.ElementTree as ElementTree

import requests
from typing import List

from castero import helpers
from castero.config import Config
from castero.episode import Episode
from castero.net import Net


class FeedError(Exception):
    """An ambiguous error while handling the feed.
    """


class FeedLoadError(FeedError):
    """A feed could not be found at the provided file, or an IO exception
    occurred when loading the file.
    """


class FeedDownloadError(FeedError):
    """A feed could not be found at the provided URL, or a request exception
    occurred when downloading the feed.
    """


class FeedParseError(FeedError):
    """The document at the provided URL could not be parsed as an XML document.
    """


class FeedStructureError(FeedError):
    """The feed at the provided URL is not a properly structured RSS document.
    """


class Feed:
    """The Feed class.

    This class uses a provided url to retrieve all data and metadata for a
    podcast feed. It creates and is a parent to all episode objects which are
    available on the feed.

    The url for the feed should point to an RSS document.
    """

    def __init__(self, url=None, file=None, **kwargs) -> None:
        """Initializes the object.

        A feed can be provided as either a url or a file, but exactly one must
        be given. Realistically, users will almost universally use a url to
        retrieve feeds from. However, having support for handling files makes
        testing easier.

        Args:
            url: (optional) the url where the feed is located
            file: (optional) the file where the feed is located
        """
        # * Don't allow providing both a url and a file, but must provide one.
        # Check that one of them is None, and that they are not both the same.
        # The second conditional can be read as checking that both variables
        # are not None.
        assert (url is None or file is None) and (url is not file)

        self._url = url
        self._file = file
        self._tree = None
        self._validated = False

        self._title = kwargs.get('title', None)
        self._description = kwargs.get('description', None)
        self._link = kwargs.get('link', None)
        self._last_build_date = kwargs.get('last_build_date', None)
        self._copyright = kwargs.get('copyright', None)

        # assume that if we have been passed the title then we have also been
        # passed everything else and that the feed is valid
        if self._title is None:
            # retrieve the feed and parse to XML document
            self._download_feed()
            # check that the XML document is a properly structured RSS feed
            self._validate_feed()
            # set this object's metadata using rss feed
            self._parse_metadata()
        else:
            self._validated = True

    def __str__(self) -> str:
        """Represent this object as a string.

        Returns:
            string: this feed's title
        """
        assert self._title is not None

        return self._title

    def _download_feed(self):
        """Parses the feed at the provided url or file into _tree.

        This method checks whether the url is valid and that there is a
        parse-able XML document at the url, but it does not check that the
        document is an RSS feed, nor whether the feed has all necessary tags.

        Raises:
            FeedParseError: unable to parse text as an XML document
            FeedDownloadError: (only when retrieving feed using url) did not
                receive an acceptable status code, or an exception occurred
                when attempting to download the page
            FeedLoadError: (only when retrieving feed using file) a feed could
                not be found at the file, or an exception occurred when
                attempting to load the file
        """
        if self._url is not None:
            # handle feed from url
            try:
                response = Net.Get(self._url)
                if response.status_code == 200:
                    try:
                        self._tree = ElementTree.fromstring(response.text)
                    except ElementTree.ParseError:
                        raise FeedParseError(
                            "Unable to parse text as an XML document")
                else:
                    raise FeedDownloadError(
                        "Did not receive an acceptable status code while"
                        " downloading the page. Expected 200, got: "
                        + str(response.status_code))
            except requests.exceptions.RequestException:
                raise FeedDownloadError(
                    "An exception occurred when attempting to download the"
                    " page")
        elif self._file is not None:
            # handle feed from file
            file = None
            try:
                file = open(self._file)
                text = file.read()
                try:
                    self._tree = ElementTree.fromstring(text)
                except ElementTree.ParseError:
                    raise FeedParseError(
                        "Unable to parse text as an XML document")
            except IOError:
                raise FeedLoadError(
                    "An exception occurred when attempting to load the file")
            finally:
                if file is not None:
                    file.close()

    def _validate_feed(self):
        """Checks that the provided XML document is a valid RSS feed.

        This method is intended to be run only when this object is being
        created in order to raise any necessary exceptions at that time.

        The conditions check are:
            - the root of the XML document is an 'rss' tag
            - the root has a 'version' attribute which equals '2.0'
            - the root has exactly one child, which is the channel tag
            - the channel tag has at least 3 children, which include a title,
              link, and description tag (in any order)
            - for each child of the channel tag which is an item tag, if any:
                - the item tag must have at least one child, which is a title
                  tag or a description tag
        See http://cyber.harvard.edu/rss/rss.html for more details.

        This method does not set this object's metadata. That is done in
        _process_feed().

        Raises:
            FeedStructureError: the XML document violates one of the conditions
        """
        assert self._tree is not None

        # root should be an rss tag
        if self._tree.tag != 'rss':
            raise FeedStructureError("XML document is not an RSS feed")

        # root should have version attribute which equals 2.0
        if 'version' in self._tree.attrib:
            if self._tree.attrib['version'] != '2.0':
                raise FeedStructureError("RSS version is not 2.0")
        else:
            raise FeedStructureError(
                "RSS feed does not have a version attribute")

        # root should have one child, which is the channel tag
        root_children = list(self._tree)
        if len(root_children) > 0:
            if len(root_children) > 1:
                raise FeedStructureError(
                    "RSS feed has too many children; expected 1, was: "
                    + str(len(root_children)))
            else:
                if root_children[0].tag != 'channel':
                    raise FeedStructureError(
                        "RSS feed does not have a channel tag as its child")
                else:
                    # channel should have at least 3 children, including a
                    # title, link, and description tag
                    channel = root_children[0]
                    channel_children = list(channel)
                    if len(channel_children) >= 3:
                        chan_title_tags = channel.findall('title')
                        chan_link_tags = channel.findall('link')
                        chan_description_tags = channel.findall('description')

                        if len(chan_title_tags) != 1:
                            raise FeedStructureError(
                                "RSS feed's channel has too many or too few"
                                " title tags; expected 1, was: "
                                + str(len(chan_title_tags)))
                        if len(chan_link_tags) != 1:
                            raise FeedStructureError(
                                "RSS feed's channel has too many or too few"
                                " link tags; expected 1, was: "
                                + str(len(chan_link_tags)))
                        if len(chan_description_tags) != 1:
                            raise FeedStructureError(
                                "RSS feed's channel has too many or too few"
                                " description tags; expected 1, was: "
                                + str(len(chan_description_tags)))

                        # if the channel has any items, each item should have
                        # at least a title or description tag
                        channel_item_tags = channel.findall('item')
                        for item in channel_item_tags:
                            if len(item.findall('title')
                                   + item.findall('description')) < 1:
                                raise FeedStructureError(
                                    "An item in the RSS feed's channel did not"
                                    " have at least one of a title or a"
                                    " description tag")
                    else:
                        raise FeedStructureError(
                            "RSS feed's channel does not have enough required"
                            " children; expected >=3, was: "
                            + str(len(channel_children)))
        else:
            raise FeedStructureError(
                "RSS feed does not have any children; expected 1 (a channel"
                " tag)")

        self._validated = True

    def _parse_metadata(self):
        """Process the RSS feed to set metadata fields.

        It is required that _validate_feed be run prior to running this method.
        """
        assert self._validated

        channel = list(self._tree)[0]
        self._title = channel.find('title').text
        self._description = channel.find('description').text
        self._link = channel.find('link').text

        last_build_date_tag = channel.find('lastBuildDate')
        copyright_tag = channel.find('copyright')

        if last_build_date_tag is not None:
            self._last_build_date = last_build_date_tag.text
        if copyright_tag is not None:
            self._copyright = copyright_tag.text

    def parse_episodes(self) -> List[Episode]:
        """Process the RSS feed to retrieve episodes.

        It is required that _validate_feed be run prior to running this method.

        Returns:
            List[Episode]: the episodes in this feed, which need to be added to
            the database
        """
        channel = list(self._tree)[0]

        episodes = []
        for item in channel.findall('item'):
            item_title = item.find('title')
            item_description = item.find('description')
            item_link = item.find('link')
            item_pubdate = item.find('pubDate')
            item_copyright = item.find('copyright')
            item_enclosure = item.find('enclosure')

            item_title_str = None
            item_description_str = None
            item_link_str = None
            item_pubdate_str = None
            item_copyright_str = None
            item_enclosure_str = None

            if item_title is not None:
                item_title_str = item_title.text
            if item_description is not None:
                item_description_str = item_description.text
            if item_link is not None:
                item_link_str = item_link.text
            if item_pubdate is not None:
                item_pubdate_str = item_pubdate.text
            if item_copyright is not None:
                item_copyright_str = item_copyright.text
            if item_enclosure is not None:
                if 'url' in item_enclosure.attrib.keys():
                    item_enclosure_str = item_enclosure.attrib['url']

            episodes.append(
                Episode(self,
                        title=item_title_str,
                        description=item_description_str,
                        link=item_link_str,
                        pubdate=item_pubdate_str,
                        copyright=item_copyright_str,
                        enclosure=item_enclosure_str
                        )
            )
        return episodes

    @property
    def validated(self) -> bool:
        """bool: whether this feed has been validated"""
        return self._validated

    @property
    def key(self) -> str:
        """str: either the url or file of the feed, whichever is set"""
        return self._url if self._url is not None else self._file

    @property
    def title(self) -> str:
        """str: the title of the feed"""
        return self._title

    @property
    def description(self) -> str:
        """str: the description of the feed"""
        return self._description

    @property
    def link(self) -> str:
        """str: the link of/for the feed"""
        return self._link

    @property
    def last_build_date(self) -> str:
        """str: the last build date of the feed"""
        result = self._last_build_date
        if result is None:
            result = "Last build date not available."
        return result

    @property
    def copyright(self) -> str:
        """str: the copyright of the feed"""
        result = self._copyright
        if result is None:
            result = "No copyright specified."
        return result

    @property
    def metadata(self) -> str:
        """str: the user-displayed metadata of the feed"""
        description = helpers.html_to_plain(self.description) if \
            helpers.is_true(Config["clean_html_descriptions"]) else \
            self.description
        description = description.replace('\n', '')

        return \
            f"\cb{self.title}\n" \
            f"{self.last_build_date}\n\n" \
            f"{self.link}\n\n" \
            f"\cbDescription:\n" \
            f"{description}\n\n" \
            f"\cbCopyright:\n" \
            f"{self.copyright}\n"



from abc import abstractmethod

import castero
from castero.config import Config
from castero.episode import Episode


class PlayerError(Exception):
    """An ambiguous error while handling the player.
    """


class PlayerDependencyError(PlayerError):
    """A dependency for playing the player was not met."""


class PlayerCreateError(PlayerError):
    """An error occurred while creating the player.
    """


class Player:
    """The Player class.

    This class is extended by players -- classes which offer methods to handle
    media operations for a specific external player (i.e. VLC, mpv).
    """

    def __init__(self, title, path, episode) -> None:
        """Initializes the object.

        Args:
            title: the title of the media (usually an episode title)
            path: a URL or file-path of a media file (usually an audio file)
            episode: the Episode which this is a player for
        """
        assert isinstance(title, str) and title != ""
        assert isinstance(path, str) and path != ""
        assert isinstance(episode, Episode) and episode is not None

        self._title = title
        self._path = path
        self._episode = episode
        self._media = None
        self._player = None
        self._duration = -1  # in milliseconds
        self._state = 0  # 0=stopped, 1=playing, 2=paused

    def __del__(self) -> None:
        if self._player is not None:
            self.stop()

    def __str__(self) -> str:
        """Represent this object as a string.

        Returns:
            string: the title of the player
        """
        return self._title

    @staticmethod
    def create_instance(available_players, title, path, episode):
        """Create an instance of an appropriate Player subclass.

        This method attempts to create a player based on the user's config
        option. If their option is not a key in available_players or
        check_dependencies() fails on the instance, we instead try to
        initialize the first working player using the order defined in
        available_players.

        Args:
            available_players: a list of implemented Player subclasses
            title: the title of the media (usually an episode title)
            path: a URL or file-path of a media file (usually an audio file)
            episode: the Episode which this is a player for

        Raises:
            PlayerDependencyError: at least one dependency per player for all
            players was not met
        """
        if Config["player"] in available_players:
            try:
                available_players[Config["player"]].check_dependencies()
                inst = available_players[Config["player"]](title, path,
                                                           episode)
                return inst
            except PlayerDependencyError:
                pass

        # Config had a bad/unsupported value; we'll instead try all implemented
        # options in order
        for av_player in sorted(available_players):
            try:
                available_players[av_player].check_dependencies()
                inst = available_players[av_player](title, path, episode)
                return inst
            except PlayerDependencyError:
                pass

        raise PlayerDependencyError("Sufficient dependencies were not met for"
                                    " any players. If you recently downloaded"
                                    " a player, you may need to reinstall %s"
                                    % castero.__title__)

    @staticmethod
    @abstractmethod
    def check_dependencies():
        """Checks whether dependencies are met for playing a player.

        Raises:
            PlayerDependencyError: a dependency was not met
        """

    @abstractmethod
    def _create_player(self) -> None:
        """Creates the player object while making sure it is a valid file.

        Checks some basic properties of the file to ensure it will play
        properly, including:
            - the media object could be parsed
            - it has a duration > 0

        Raises:
            PlayerCreateError: the player object could not be created
        """

    @abstractmethod
    def play(self) -> None:
        """Plays the media.
        """

    @abstractmethod
    def stop(self) -> None:
        """Stops the media.
        """

    @abstractmethod
    def pause(self) -> None:
        """Pauses the media.
        """

    @abstractmethod
    def seek(self, direction, amount) -> None:
        """Seek forward or backward in the media.

        Args:
            direction: 1 to seek forward, -1 to seek backward
            amount: the amount of seconds to seek
        """

    @property
    def state(self) -> int:
        """int: the state of the player"""
        return self._state

    @property
    def title(self) -> str:
        """str: the title of the player"""
        return self._title

    @property
    def episode(self) -> Episode:
        """Episode: the Episode which this player has the media for"""
        return self._episode

    @property
    @abstractmethod
    def duration(self) -> int:
        """int: the duration of the player, in ms"""

    @property
    @abstractmethod
    def time(self) -> int:
        """int: the current time of the player, in ms"""

    @property
    @abstractmethod
    def time_str(self) -> str:
        """str: the formatted time and duration of the player"""



import threading

from castero.episode import Episode


class DownloadQueue:
    """The DownloadQueue class.

    A FIFO ordered queue for handling episode downloads.
    """

    def __init__(self, display=None) -> None:
        """Initializes the object.
        """
        self._episodes = []
        self._display = display

    def next(self) -> None:
        """Proceed to the next episode in the queue.
        """
        if len(self._episodes) > 0:
            self._episodes.pop(0)
            self.start()

    def add(self, episode) -> None:
        """Adds an episode to the end of the queue.
        """
        assert isinstance(episode, Episode)

        if episode not in self._episodes:
            self._episodes.append(episode)

    def start(self) -> None:
        """Start downloading the first episode in the queue.
        """
        if self.first is not None:
            self.first.download(self, self._display)

    def update(self) -> None:
        """Checks the status of the current download.
        """
        # if nothing is downloading, start downloading the first episode
        found_downloading = False
        for thread in threading.enumerate():
            if thread.getName().startswith("download"):
                found_downloading = True
        if not found_downloading and len(self._episodes) > 0:
            self.start()

    @property
    def first(self) -> Episode:
        """Episode: the first episode in the queue"""
        result = None
        if len(self._episodes) > 0:
            result = self._episodes[0]
        return result

    @property
    def length(self) -> int:
        """int: the length of the queue"""
        return len(self._episodes)



__title__ = "castero"
__description__ = "A TUI podcast client for the terminal"
__keywords__ = "podcast commandline terminal tui curses"
__author__ = "Jake Robertson"
__author_email__ = "jake@faltro.com"
__version__ = "0.5.0"
__date__ = "2019-03-19"
__copyright__ = "Copyright (c) 2018 Jake Robertson"
__license__ = "MIT License"
__url__ = "https://github.com/xgi/castero"

__help__ = """\
%s help
============

Version: %s
Updated: %s
Maintainer: %s <%s>
License: %s
URL: %s

Commands
    {key_help} - show this help screen
    {key_exit} - exit the client
    {key_add_feed} - add a feed
    {key_delete} - delete the selected feed
    {key_reload} - reload/refresh feeds
    {key_save} - save episode for offline playback
    {key_up} - navigate up in menus
    {key_right} - navigate right in menus
    {key_down} - navigate down in menus
    {key_left} - navigate left in menus
    {key_scroll_up} - scroll up in menus
    {key_scroll_down} - scroll down in menus
    {key_play_selected} - play selected feed/episode
    {key_add_selected} - add selected feed/episode to queue
    {key_clear} - clear the queue
    {key_next} - go to the next episode in the queue
    {key_invert} - invert the order of the menu
    {key_mark_played} - mark episode as played/unplayed
    {key_pause_play|key_pause_play_alt} - pause/play the current episode
    {key_seek_forward|key_seek_forward_alt} - seek forward
    {key_seek_backward|key_seek_backward_alt} - seek backward
    {1-3} - change between client layouts
""" % (
    __title__,
    __version__,
    __date__,
    __author__,
    __author_email__,
    __license__,
    __url__
)



import curses

from castero import helpers
from castero.config import Config
from castero.feed import Feed
from castero.menu import Menu


class QueueMenu(Menu):
    def __init__(self, window, source, child=None, active=False) -> None:
        super().__init__(window, source, child=child, active=active)

    def __len__(self) -> int:
        return self._source.length

    def _items(self):
        return [
            {
                'attr': curses.A_NORMAL,
                'text': str(player)
            }
            for player in self._source
        ]

    def item(self) -> Feed:
        if self._source.length == 0:
            return None
        
        return self._source[self._selected]

    def metadata(self):
        player = self.item()
        if player is None:
            return ""

        return player.episode.metadata

    def update_items(self, obj):
        super().update_items(obj)

    def update_child(self):
        pass

    def invert(self):
        pass



import curses
from abc import ABC, abstractmethod, abstractproperty
from typing import List


class Menu(ABC):
    """The Menu class.

    This class is used to display interactable menus. It displays a list of
    items to its window and allows the user to cycle through them.

    This class does not handle user input -- that is done in the Display class.
    Methods in that class simply call appropriate methods here in response to
    user input in order to change the state of the menu.
    """

    @abstractmethod
    def __init__(self, window, source, child=None, active=False) -> None:
        """Initializes the object.

        Args:
            window: the curses.window which this menu is placed on
            items: a 2D array where rows represent indices of the parent menu
            child: (optional, default None) the submenu of this menu
            active: (optional, default False) whether this menu is active
        """
        assert window.getmaxyx()[0] >= 3
        assert child is None or isinstance(child, Menu)
        assert isinstance(active, bool)

        self._window = window
        self._source = source
        self._child = child
        self._active = active
        self._selected = 0
        self._display_start_y = 2
        self._top_index = 0
        self._max_displayed_items = self._window.getmaxyx()[0] - 4
        self._inverted = False

        if child is not None:
            self.update_child()

    @abstractmethod
    def __len__(self) -> int:
        """int: the number of items in the menu"""

    @abstractproperty
    @property
    def _items(self) -> List[str]:
        """List[str]: the current items in this menu"""

    @abstractproperty
    @property
    def item(self):
        """the selected item"""

    @abstractproperty
    @property
    def metadata(self) -> str:
        """str: metadata for the selected item"""

    @abstractmethod
    def update_items(self, obj) -> None:
        """Called by the parent menu (if we have one) to update our items.

        Args:
            obj: an object of some type understood by the specific
            implementation of this menu
        """

    @abstractmethod
    def update_child(self) -> None:
        """Implementation-specific method to update our child, if we have one.

        This method calls the child's update_items with an
        implementation-specific object understood by the child.
        """

    @abstractmethod
    def invert(self) -> None:
        """Invert the menu order.

        Inversion is not just a visual effect -- the contents of the items list
        must also be reversed.
        """
        self._inverted = not self._inverted

    def _pad_text(self, text) -> str:
        """Pads an item string with spaces to be the full length of the menu.

        Note that this does not create a string the entire length of the
        window, it only makes it as large as the menu. Since the window has a
        border, the string will be 1 column shorter than the window.

        Args:
            item: the item text as a string

        Returns:
            str: a string the length of the menu, with the left justified item
        """
        max_width = self._window.getmaxyx()[1] - 1
        return text.ljust(max_width)[:max_width]

    def _draw_item(self, index, position) -> None:
        """Draws an item's text on the window.

        This method applies the appropriate color pair to the item: 2 if the
        item is selected and the window is active, 3 if the item is selected
        but the window is not active, and 1 otherwise.

        Args:
            index: the index of the item
            position: the y-position to draw the item, without accounting
                for _display_start_y
        """
        item = self._items()[index]
        item_str = self._pad_text(item['text'])

        attr = curses.color_pair(1)
        if index == self._selected:
            if self._active:
                attr = curses.color_pair(2)
            else:
                attr = curses.color_pair(3)
        else:
            attr = attr | item['attr']

        self._window.addstr(self._display_start_y + position, 0,
                            item_str, attr)

    def _sanitize(self) -> None:
        """Sanitizes _selected and _top_index.

        Checks that _selected and _top_index are valid (inside all boundaries),
        setting them to appropriate extremes if they are not.
        """
        num_my_items = len(self._items())

        # _selected cannot be outside range of items
        if self._selected < 0:
            self._selected = 0
        if self._selected > num_my_items - 1:
            self._selected = num_my_items - 1

        # if there is no next page, then the current page should be as full
        # as possible
        if self._top_index + self._max_displayed_items > num_my_items:
            self._top_index = num_my_items - self._max_displayed_items

        # _top_index cannot be outside range of items
        if self._top_index > num_my_items - 1:
            self._top_index = num_my_items - 1
        if self._top_index < 0:
            self._top_index = 0

    def display(self) -> None:
        """Draw all visible items on this menu to the window.

        Visible items are items with an index greater than or equal to
        _top_index but less than _max_displayed_items greater than _top_index.
        That is, all items that can fit on the screen starting from _top_index.
        """
        position = 0
        for i in range(self._top_index,
                       self._top_index + self._max_displayed_items):
            if i <= len(self._items()) - 1:
                self._draw_item(i, position)
                position += 1

        # fill unused rows with blank lines
        for y in range(self._display_start_y + position,
                       self._display_start_y + self._max_displayed_items):
            self._window.addstr(y, 0, self._pad_text(""))

    def set_active(self, active) -> None:
        """Sets whether this menu is active.

        The only effect this method has on this object is whether or not to
        display the selected item as yellow or as grayed-out. Any movement
        operations could still be run, but doing so would confuse the user
        assuming _active has been properly set.

        Args:
            active: whether this menu is active or not
        """
        assert isinstance(active, bool)

        self._active = active

    def move(self, direction) -> None:
        """Change the selected item to an adjacent item.

        Args:
            direction: 1 to move up, -1 to move down
        """
        assert direction == 1 or direction == -1

        self._selected -= direction

        if self._selected < self._top_index:
            # the cursor went above the menu
            self._top_index -= 1
        elif self._selected >= self._top_index + self._max_displayed_items:
            # the cursor went below the menu
            self._top_index += 1

        self._sanitize()
        if self._child is not None:
            self.update_child()

    def move_page(self, direction) -> None:
        """Change the selected item to the next "page".

        Effectively the same as moving _max_displayed_items times.

        We always try to make the menu as full as possible -- if the movement
        would leave us with only a few items on the screen, we instead reset
        _top_index to make the screen as full as possible. Therefore we can
        assume that if there are enough items in the menu to fill the screen,
        the menu will *always* fill the screen.

        Args:
            direction: 1 to move up, -1 to move down
        """
        assert direction == 1 or direction == -1

        if direction == 1:
            self._selected -= self._max_displayed_items
            self._top_index -= self._max_displayed_items
        elif direction == -1:
            self._selected += self._max_displayed_items
            self._top_index += self._max_displayed_items

        self._sanitize()
        if self._child is not None:
            self.update_child()

    @property
    def selected_index(self) -> int:
        """int: the current selected index of the menu"""
        return self._selected



import setuptools

import castero

install_requires = [
    'requests',
    'cjkwrap',
    'beautifulsoup4',
    'python-vlc',
    'python-mpv'
]

tests_require = [
    'pytest',
    'coverage',
    'codacy-coverage'
]

extras_require = {
    'test': tests_require
}


def long_description():
    with open("README.md") as readme:
        return readme.read()


setuptools.setup(
    name=castero.__title__,
    version=castero.__version__,
    description=castero.__description__,
    long_description=long_description(),
    long_description_content_type='text/markdown',
    keywords=castero.__keywords__,
    url=castero.__url__,
    author=castero.__author__,
    author_email=castero.__author_email__,
    license=castero.__license__,
    packages=[
        'castero', 'castero.perspectives', 'castero.players', 'castero.menus'
    ],
    package_data={
        'castero': ['templates/*'],
    },
    python_requires='>=3',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    entry_points={'console_scripts': ['castero=castero.__main__:main']},
    classifiers=[
        'Intended Audience :: End Users/Desktop',
        'Environment :: Console :: Curses',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Topic :: Terminals'
    ],
)



import curses

from castero import helpers
from castero.config import Config
from castero.feed import Feed
from castero.menu import Menu
from castero.menus.episodemenu import EpisodeMenu


class FeedMenu(Menu):
    def __init__(self, window, source, child=None, active=False) -> None:
        assert child is not None and isinstance(child, EpisodeMenu)

        self._feeds = []

        super().__init__(window, source, child=child, active=active)

    def __len__(self) -> int:
        return len(self._feeds)

    def _items(self):
        return [
            {
                'attr': curses.A_NORMAL,
                'text': str(feed)
            }
            for feed in self._feeds
        ]

    def item(self) -> Feed:
        if len(self._feeds) == 0:
            return None
        
        return self._feeds[self._selected]

    def metadata(self):
        feed = self.item()
        if feed is None:
            return ""

        return feed.metadata

    def update_items(self, obj):
        super().update_items(obj)

        self._feeds = self._source.feeds()
        if self._inverted:
            self._feeds.reverse()

        self._sanitize()

    def update_child(self):
        if len(self._feeds) == 0:
            self.update_items(None)
        else:
            self._child.update_items(self._feeds[self._selected])

    def invert(self):
        super().invert()

        self.update_items(None)



import time

from castero.player import Player, PlayerDependencyError


class MPVPlayer(Player):
    """The MPVPlayer class.
    """
    NAME = "mpv"

    def __init__(self, title, path, episode) -> None:
        """Initializes the object.

        Overrides method from Player; see documentation in that class.
        """
        super().__init__(title, path, episode)

        import mpv
        self.mpv = mpv

    @staticmethod
    def check_dependencies():
        """Checks whether dependencies are met for playing a player.

        Overrides method from Player; see documentation in that class.
        """
        try:
            import mpv
            mpv.MPV()
        except (ImportError, NameError, OSError, AttributeError,
                ModuleNotFoundError):
            raise PlayerDependencyError(
                "Dependency mpv not found, which is required for playing"
                " media files"
            )

    def _create_player(self) -> None:
        """Creates the player object while making sure it is a valid file.

        Overrides method from Player; see documentation in that class.
        """
        self._player = self.mpv.MPV()
        self._player.vid = False
        self._player.pause = False

        self._player.play(self._path)

        self._duration = 5

    def play(self) -> None:
        """Plays the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is None:
            self._create_player()

        self._player.pause = False
        self._state = 1

    def stop(self) -> None:
        """Stops the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is not None:
            self._player.terminate()
            self._state = 0

    def pause(self) -> None:
        """Pauses the media.

        Overrides method from Player; see documentation in that class.
        """
        if self._player is not None:
            self._player.pause = True
            self._state = 2

    def seek(self, direction, amount) -> None:
        """Seek forward or backward in the media.

        Overrides method from Player; see documentation in that class.
        """
        assert direction == 1 or direction == -1
        if self._player is not None:
            self._player.seek(direction * amount)

    @property
    def duration(self) -> int:
        """int: the duration of the player"""
        result = 0
        if self._player is not None:
            d = self._player.duration
            result = 5000 if d is None else d * 1000
        return result

    @property
    def time(self) -> int:
        """int: the current time of the player"""
        if self._player is not None:
            t = self._player.time_pos
            return 0 if t is None else t * 1000

    @property
    def time_str(self) -> str:
        """str: the formatted time and duration of the player"""
        result = "00:00:00/00:00:00"
        if self._player is not None:
            time_seconds = int(self.time / 1000)
            length_seconds = int(self.duration / 1000)
            t = time.strftime('%H:%M:%S', time.gmtime(time_seconds))
            d = time.strftime('%H:%M:%S', time.gmtime(length_seconds))
            result = "%s/%s" % (t, d)
        return result
